# -*- coding: UTF-8 -*-
# Injection Wizard
# --An application for injecting traffic in WEP protected Wi-Fi networks--
# Version: 1.0
# Copyright (C) 2010 Fernando Pablo Romero Navarro <fernpromero@gmail.com>
#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import socket
import time


class CommServ(object):
     
    def __init__(self, timeout):
        self.rdtimeout=timeout

    def readmsg(self):
        pass

    def writemsg(self, msg):
        pass

    def release(self):
        pass

class NetCommServ(CommServ):

    MAX_BYTES=4096
    MAX_RETRIES=8
    EOL_SEQ="\n"
    ESC_SEQ="\\"

    def __init__(self, cntdsock, timeout):
        CommServ.__init__(self, timeout)
        self.sock=cntdsock
        self.sock.settimeout(timeout)
        self.buf=None
        
    def readmsg(self):
        
        eol=False
        buf=""

        try:
            while(not eol and len(buf) < self.MAX_BYTES):
                data=self.sock.recv(self.MAX_BYTES)

                if(len(data)==0):
                    raise socket.error
                
                eolpos=data.find(self.EOL_SEQ)

                while eolpos>-1 and not eol:
                    if(data[eolpos-1]==self.ESC_SEQ):
                        data=data[:eolpos-1]+data[eolpos:]
                        eolpos=data.find(self.EOL_SEQ, eolpos)
                    else:
                        eol=True
                        
                if(eol):
                   buf=buf+data[:eolpos]
                else:
                   buf=buf+data

            if(eol):
                if(self.buf is not None):
                    buf=self.buf+buf
                    self.buf=None                

                if(len(data)-1 > eolpos):
                    self.buf=data[eolpos+1:]

                return buf

            else:
                if(self.buf is not None):
                    buf=self.buf+buf
                    self.buf=None
                    
                return buf[:self.MAX_BYTES]

        except socket.timeout:  

            if len(buf)>0:
                if self.buf is None:
                    self.buf=buf
                    raise socket.timeout
                else:
                    self.buf=self.buf+buf
                    if(len(self.buf) < self.MAX_BYTES):
                        raise socket.timeout
                    else:
                        tmp=self.buf[:self.MAX_BYTES]
                        self.buf=None
                        return tmp
            else:
                raise


    def writemsg(self, msg):
        
        if(msg.find(self.EOL_SEQ)>-1):
            msg=msg.replace(self.EOL_SEQ, self.ESC_SEQ+self.EOL_SEQ)
        
        self.sock.sendall(msg+self.EOL_SEQ)
        


    def release(self):
        self.buf=None
        self.sock.close()

        

class DummyCommServ(CommServ):

    def __init__(self, timeout):
        CommServ.__init__(self, timeout)

    def readmsg(self):
        time.sleep(self.rdtimeout)
        raise socket.timeout

    def writemsg(self, msg):
        print repr(msg)